@@ -38,7 +38,6 @@
 #include "qquick3dmodel_p.h"
 #include "qquick3drenderstats_p.h"
 #include <QtQuick3DRuntimeRender/private/qssgrenderlayer_p.h>
-#include <QOpenGLFunctions>
 
 #include <qsgtextureprovider.h>
 #include <QSGSimpleTextureNode>
@@ -56,36 +55,41 @@ QT_BEGIN_NAMESPACE
     \inqmlmodule QtQuick3D
     \brief Provides a viewport on which to render a 3D scene.
 
-    View3D provides a 2D surface for 3D content to be rendered to. Before 3D
-    content can be displayed in a Qt Quick scene, it must first be flattend.
+    View3D provides a 2D surface on which a 3D scene can be rendered. This
+    surface is a Qt Quick \l Item and can be placed in a Qt Quick scene.
 
-    There are two ways to define a 3D scene for View3D to view.  The first
-    and easiest is to just define a higharchy of \l Node based items as
-    children of the View3D. This becomes the implicit scene of the viewport.
+    There are two ways to define the 3D scene that is visualized on the View3D:
+    If you define a hierarchy of \l{Node}{Node-based} items as children of
+    the View3D directly, then this will become the implicit scene of the View3D.
 
-    It is also possible to reference an existing scene by using the
-    \l importScene property of the scene you want to render.
-    This scene does not have to exist as a child of the current View3D.
+    It is also possible to reference an existing scene by using the \l importScene
+    property and setting it to the root \l Node of the scene you want to visualize.
+    This \l Node does not have to be an ancestor of the View3D, and you can have
+    multiple View3Ds that import the same scene.
 
-    There is also a combination approach where you define both a scene with
-    children nodes, as well as define a scene that is being referenced. In this
-    case you can treat the referenced scene as a sibling of the child scene.
+    This is demonstrated in \l {Qt Quick 3D - View3D example}{View3D example}.
 
-    This is demonstrated in \l {Qt Quick 3D - View3D example}{View3D example}
+    If the View3D both has child \l{Node}{Nodes} and the \l importScene property is
+    set simultaneously, then both scenes will be rendered as if they were sibling
+    subtrees in the same scene.
 
-    To control how a scene is rendered, it is necessary to define a
-    \l SceneEnvironment to the \l environment property.
+    To control how a scene is rendered, you can set the \l environment property.
+    The type \l SceneEnvironment has a number of visual properties that can be
+    adjusted, such as background color, tone mapping, anti-aliasing and more.
 
-    To project a 3D scene to a 2D viewport, it is necessary to view the scene
-    from a camera. If a scene has more than one camera it is possible to set
-    which camera is used to render the scene to this viewport by setting the
-    \l camera property.
+    In addition, in order for anything to be rendered in the View3D, the scene
+    needs a \l Camera. If there is only a single \l Camera in the scene, then
+    this will automatically be picked. Otherwise, the \l camera property can
+    be used to select the camera. The \l Camera decides which parts of the scene
+    are visible, and how they are projected onto the 2D surface.
 
-    It is also possible to define where the 3D scene is rendered to using
-    the \l renderMode property. This can be necessary for performance
-    reasons on certain platforms where it is expensive to render to
-    intermediate offscreen surfaces. There are certain tradeoffs to rendering
-    directly to the window though, so this is not the default.
+    By default, the 3D scene will first be rendered into an off-screen buffer and
+    then composited with the rest of the Qt Quick scene when it is done. This provides
+    the maximum level of compatibility, but may have performance implications on some
+    graphics hardware. If this is the case, the \l renderMode property can be used to
+    switch how the View3D is rendered into the window.
+
+    \sa {Qt Quick 3D - View3D example}
 */
 
 QQuick3DViewport::QQuick3DViewport(QQuickItem *parent)
@@ -96,9 +100,10 @@ QQuick3DViewport::QQuick3DViewport(QQuickItem *parent)
     m_sceneRoot = new QQuick3DSceneRootNode(this);
     m_environment = new QQuick3DSceneEnvironment(m_sceneRoot);
     m_renderStats = new QQuick3DRenderStats(m_sceneRoot);
-    QSharedPointer<QQuick3DSceneManager> sceneManager(new QQuick3DSceneManager(m_sceneRoot));
-    QQuick3DObjectPrivate::get(m_sceneRoot)->refSceneManager(sceneManager);
-    connect(QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager.data(), &QQuick3DSceneManager::needsUpdate,
+    QQuick3DSceneManager *sceneManager = new QQuick3DSceneManager(m_sceneRoot);
+    QQuick3DObjectPrivate::get(m_sceneRoot)->refSceneManager(*sceneManager);
+    Q_ASSERT(sceneManager == QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager);
+    connect(sceneManager, &QQuick3DSceneManager::needsUpdate,
             this, &QQuickItem::update);
 }
 
@@ -106,11 +111,20 @@ QQuick3DViewport::~QQuick3DViewport()
 {
     for (const auto &connection : qAsConst(m_connections))
         disconnect(connection);
-    // Do not delete scenemanager along with sceneroot
-    auto &sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
+    auto sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
     if (sceneManager)
         sceneManager->setParent(nullptr);
+
     delete m_sceneRoot;
+    m_sceneRoot = nullptr;
+
+    // We need to defere the deletion of the scene manager or items might
+    // try to deref using the now destroyed scene manaager.
+    connect(this, &QObject::destroyed, [sceneManager](QObject *) { delete sceneManager; });
+    sceneManager = nullptr;
+
+    // m_directRenderer must be destroyed on the render thread at the proper time, not here.
+    // That's handled in releaseResources() + upon sceneGraphInvalidated
 }
 
 static void ssgn_append(QQmlListProperty<QObject> *property, QObject *obj)
@@ -128,7 +142,7 @@ static void ssgn_append(QQmlListProperty<QObject> *property, QObject *obj)
     }
 }
 
-static int ssgn_count(QQmlListProperty<QObject> *property)
+static qsizetype ssgn_count(QQmlListProperty<QObject> *property)
 {
     QQuick3DViewport *view3d = static_cast<QQuick3DViewport *>(property->object);
     if (!view3d || !view3d->scene() || !QQuick3DObjectPrivate::get(view3d->scene())->data().count)
@@ -137,7 +151,7 @@ static int ssgn_count(QQmlListProperty<QObject> *property)
     return itemProperty.count(&itemProperty);
 }
 
-static QObject *ssgn_at(QQmlListProperty<QObject> *property, int i)
+static QObject *ssgn_at(QQmlListProperty<QObject> *property, qsizetype i)
 {
     QQuick3DViewport *view3d = static_cast<QQuick3DViewport *>(property->object);
     QQmlListProperty<QObject> itemProperty = QQuick3DObjectPrivate::get(view3d->scene())->data();
@@ -165,12 +179,8 @@ QQmlListProperty<QObject> QQuick3DViewport::data()
 /*!
     \qmlproperty QtQuick3D::Camera QtQuick3D::View3D::camera
 
-    This property specifies which camera is used to render the scene. It is
-    possible for this value to be undefined, in which case the first enabled
-    camera in the scene will be used instead.
-
-    If it is desired to not render anything in the scene, then make sure all
-    cameras are disabled.
+    This property specifies which \l Camera is used to render the scene. If this
+    property is not set, then the first enabled camera in the scene will be used.
 
     \sa PerspectiveCamera, OrthographicCamera, FrustumCamera, CustomCamera
 */
@@ -193,8 +203,9 @@ QQuick3DSceneEnvironment *QQuick3DViewport::environment() const
 
 /*!
     \qmlproperty QtQuick3D::Node QtQuick3D::View3D::scene
+    \readonly
 
-    Holds the root scene of the View3D.
+    Holds the root \l Node of the View3D's scene.
 
     \sa importScene
 */
@@ -206,10 +217,12 @@ QQuick3DNode *QQuick3DViewport::scene() const
 /*!
     \qmlproperty QtQuick3D::Node QtQuick3D::View3D::importScene
 
-    This property defines the reference node of the scene to render to the
-    viewport. The node does not have to be a child of the View3D.
-    This referenced node becomes sibling with possible child nodes of View3D.
-    \note This property can only be set once, not removed or changed later.
+    This property defines the reference node of the scene to render to the viewport.
+    The node does not have to be a child of the View3D. This referenced node becomes
+    a sibling with child nodes of View3D, if there are any.
+
+    \note This property can only be set once, and subsequent changes will have no
+    effect.
 
     \sa Node
 */
@@ -221,14 +234,45 @@ QQuick3DNode *QQuick3DViewport::importScene() const
 /*!
     \qmlproperty enumeration QtQuick3D::View3D::renderMode
 
-    This property determines how the scene is rendered to the viewport.
+    This property determines how the View3D is combined with the other parts of the
+    Qt Quick scene.
 
-    \value View3D.Offscreen Scene is rendered to a texture. Comes with no limitations.
-    \value View3D.Underlay Scene is rendered directly to the window before Qt Quick is rendered.
-    \value View3D.Overlay Scene is rendered directly to the window after Qt Quick is rendered.
-    \value View3D.Inline Scene is rendered to the current render target using QSGRenderNode.
+    By default, the scene will be rendered into an off-screen buffer as an intermediate
+    step. This off-screen buffer is then rendered into the window (or render target) like any
+    other Qt Quick \l Item.
 
-    The default mode is \c View3D.Offscreen as this is the offers the best compatibility.
+    For most users, there will be no need to change the render mode, and this property can
+    safely be ignored. But on some graphics hardware, the use of an off-screen buffer can be
+    a performance bottleneck. If this is the case, it might be worth experimenting with other
+    modes.
+
+    \value View3D.Offscreen The scene is rendered into an off-screen buffer as an intermediate
+    step. This off-screen buffer is then composited with the rest of the Qt Quick scene.
+
+    \value View3D.Underlay The scene is rendered directly to the window before the rest of
+    the Qt Quick scene is rendered. With this mode, the View3D cannot be placed on top of
+    other Qt Quick items.
+
+    \value View3D.Overlay The scene is rendered directly to the window after Qt Quick is
+    rendered.  With this mode, the View3D will always be on top of other Qt Quick items.
+
+    \value View3D.Inline The View3D's scene graph is embedded into the main scene graph,
+    and the same ordering semantics are applied as to any other Qt Quick \l Item. As this
+    mode can lead to subtle issues, depending on the contents of the scene, due to
+    injecting depth-based 3D content into a 2D scene graph, it is not recommended to be
+    used, unless a specific need arises.
+
+    The default is \c{View3D.Offscreen}.
+
+    \note When changing the render mode, it is important to note that \c{View3D.Offscreen} (the
+    default) is the only mode which guarantees perfect graphics fidelity. The other modes
+    all have limitations that can cause visual glitches, so it is important to check that
+    the visual output still looks correct when changing this property.
+
+    \note When using the Underlay, Overlay, or Inline modes, it can be useful, and in some
+    cases, required, to disable the Qt Quick scene graph's depth buffer writes via
+    QQuickGraphicsConfiguration::setDepthBufferFor2D() before showing the QQuickWindow or
+    QQuickView hosting the View3D item.
 */
 QQuick3DViewport::RenderMode QQuick3DViewport::renderMode() const
 {
@@ -237,8 +281,9 @@ QQuick3DViewport::RenderMode QQuick3DViewport::renderMode() const
 
 /*!
     \qmlproperty QtQuick3D::RenderStats QtQuick3D::View3D::renderStats
+    \readonly
 
-    Accessor to \l {RenderStats}, which can be used to gain information of
+    This property provides statistics about the rendering of a frame, such as
     \l {RenderStats::fps}{fps}, \l {RenderStats::frameTime}{frameTime},
     \l {RenderStats::renderTime}{renderTime}, \l {RenderStats::syncTime}{syncTime},
     and \l {RenderStats::maxFrameTime}{maxFrameTime}.
@@ -248,113 +293,6 @@ QQuick3DRenderStats *QQuick3DViewport::renderStats() const
     return m_renderStats;
 }
 
-void QQuick3DViewport::setShaderCacheFile(const QUrl &shaderCacheFile)
-{
-    m_shaderCacheFile = shaderCacheFile;
-}
-
-QUrl QQuick3DViewport::shaderCacheFile()
-{
-    return m_shaderCacheFile;
-}
-
-void QQuick3DViewport::readShaderCache()
-{
-    QByteArray error;
-    if (!shaderCacheFile().isEmpty()) {
-        QFile file(QQmlFile::urlToLocalFileOrQrc(shaderCacheFile()));
-        if (file.open(QIODevice::ReadOnly))
-            m_shaderCacheData = qUncompress(file.readAll());
-
-        if (m_shaderCacheData.isEmpty()) {
-            error = QByteArrayLiteral("Failed to read or uncompress shader cache: ");
-            error.append(shaderCacheFile().toString().toUtf8());
-            error.append(" ");
-            error.append(file.errorString().toUtf8());
-        }
-    } else if (!m_shaderCacheImport.isEmpty()) {
-        m_shaderCacheData = qUncompress(m_shaderCacheImport);
-        if (m_shaderCacheData.isEmpty())
-            error = QByteArrayLiteral("Failed uncompress shader cache.");
-    }
-
-    if (!error.isEmpty())
-        Q_EMIT shaderCacheLoadErrors(error);
-}
-
-void QQuick3DViewport::writeShaderCache(const QUrl &shaderCacheFile)
-{
-    if (m_shaderCacheData.isEmpty()) {
-        Q_EMIT shaderCacheExported(false);
-        return; // Warning is already printed by export function
-    }
-    const QString filePath = shaderCacheFile.toLocalFile();
-    if (filePath.isEmpty()) {
-        qWarning() << __FUNCTION__ << "Warning: Invalid filename: " << shaderCacheFile;
-        Q_EMIT shaderCacheExported(false);
-        return;
-    }
-    QSaveFile file(filePath);
-    QFileInfo(filePath).dir().mkpath(QStringLiteral("."));
-    bool success = false;
-    if (file.open(QIODevice::WriteOnly) && file.write(m_shaderCacheData) != -1) {
-        file.commit();
-        success = true;
-    } else {
-        qWarning() << __FUNCTION__ << "Warning: Failed to write shader cache:"
-                   << shaderCacheFile << file.errorString();
-    }
-    Q_EMIT shaderCacheExported(success);
-}
-
-void QQuick3DViewport::doExportShaderCache()
-{
-    if (m_exportShaderCacheRequested) {
-        if (!QOpenGLContext::currentContext()) {
-            qWarning () << "Unable to export shader cache. No current context.";
-            m_exportShaderCacheRequested = false;
-            Q_EMIT shaderCacheExported(false);
-            return;
-        }
-        auto rci = QSSGRenderContextInterface::getRenderContextInterface(quintptr(window()));
-        if (rci) {
-            m_shaderCacheData = rci->shaderCache()->exportShaderCache(m_binaryShaders);
-            if (m_shaderCacheData.size()) {
-                m_shaderCacheData = qCompress(m_shaderCacheData, m_compressionLevel);
-                writeShaderCache(m_exportShaderCacheFile);
-            } else {
-                Q_EMIT shaderCacheExported(false);
-            }
-        }
-        m_exportShaderCacheRequested = false;
-    }
-}
-
-void QQuick3DViewport::doImportShaderCache()
-{
-    readShaderCache();
-    if (!m_shaderCacheData.isNull()) {
-        QByteArray error;
-        auto rci = QSSGRenderContextInterface::getRenderContextInterface(quintptr(window()));
-        rci->shaderCache()->importShaderCache(m_shaderCacheData, error);
-        if (!error.isEmpty())
-            Q_EMIT shaderCacheLoadErrors(error);
-    }
-}
-
-void QQuick3DViewport::exportShaderCache(const QUrl &shaderCacheFile, bool binaryShaders,
-                                         int compressionLevel)
-{
-    if (m_exportShaderCacheRequested) {
-        qWarning () << "Export shader cache already requested";
-        return;
-    }
-    m_exportShaderCacheFile = shaderCacheFile;
-    m_binaryShaders = binaryShaders;
-    m_compressionLevel = compressionLevel;
-    m_exportShaderCacheRequested = true;
-}
-
 QQuick3DSceneRenderer *QQuick3DViewport::createRenderer() const
 {
     return new QQuick3DSceneRenderer(this->window());
@@ -382,10 +320,11 @@ QSGTextureProvider *QQuick3DViewport::textureProvider() const
         return nullptr;
 
     QQuickWindow *w = window();
-    if (!w || !w->openglContext() || QThread::currentThread() != w->openglContext()->thread()) {
+    if (!w) {
         qWarning("QSSGView3D::textureProvider: can only be queried on the rendering thread of an exposed window");
         return nullptr;
     }
+
     if (!m_node)
         m_node = new SGFramebufferObjectNode;
     return m_node;
@@ -416,9 +355,9 @@ void QQuick3DViewport::cleanupDirectRenderer()
     m_directRenderer = nullptr;
 }
 
-void QQuick3DViewport::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
+void QQuick3DViewport::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
-    QQuickItem::geometryChanged(newGeometry, oldGeometry);
+    QQuickItem::geometryChange(newGeometry, oldGeometry);
 
     if (newGeometry.size() != oldGeometry.size())
         update();
@@ -426,6 +365,15 @@ void QQuick3DViewport::geometryChanged(const QRectF &newGeometry, const QRectF &
 
 QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePaintNodeData *)
 {
+    // ### Maybe don't check this every time, only when the window changes
+    QSGRendererInterface *rif = window()->rendererInterface();
+    const bool isRhi = QSGRendererInterface::isApiRhiBased(rif->graphicsApi());
+    if (!isRhi) {
+        // We only support using the RHI API, so bail out gracefully
+        // ### maybe return a node tree that displays a warning that this isn't supported
+        return nullptr;
+    }
+
     // When changing render modes
     if (m_renderModeDirty) {
         if (node) {
@@ -438,13 +386,10 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
             delete m_directRenderer;
             m_directRenderer = nullptr;
         }
-        updateClearBeforeRendering();
     }
 
     m_renderModeDirty = false;
 
-    doExportShaderCache();
-
     if (m_renderMode == Offscreen) {
         SGFramebufferObjectNode *n = static_cast<SGFramebufferObjectNode *>(node);
 
@@ -457,10 +402,9 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
         if (!n->renderer) {
             n->window = window();
             n->renderer = createRenderer();
-            n->renderer->data = n;
+            n->renderer->fboNode = n;
             n->quickFbo = this;
             connect(window(), SIGNAL(screenChanged(QScreen*)), n, SLOT(handleScreenChange()));
-            doImportShaderCache();
         }
         QSize minFboSize = QQuickItemPrivate::get(this)->sceneGraphContext()->minimumFBOSize();
         QSize desiredFboSize(qMax<int>(minFboSize.width(), width()),
@@ -469,14 +413,16 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
         n->devicePixelRatio = window()->effectiveDevicePixelRatio();
         desiredFboSize *= n->devicePixelRatio;
 
-        n->setTextureCoordinatesTransform(QSGSimpleTextureNode::MirrorVertically);
         n->setFiltering(smooth() ? QSGTexture::Linear : QSGTexture::Nearest);
         n->setRect(0, 0, width(), height());
         if (checkIsVisible() && isComponentComplete()) {
-            n->renderer->synchronize(this, desiredFboSize);
+            n->renderer->synchronize(this, desiredFboSize, n->devicePixelRatio);
+            if (n->renderer->m_textureNeedsFlip)
+                n->setTextureCoordinatesTransform(QSGSimpleTextureNode::MirrorVertically);
             updateDynamicTextures();
             n->scheduleRender();
         }
+
         return n;
     } else if (m_renderMode == Underlay) {
         setupDirectRenderer(Underlay);
@@ -484,31 +430,32 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
     } else if (m_renderMode == Overlay) {
         setupDirectRenderer(Overlay);
         return node; // node should be nullptr
-    } else {
-        // Render Node
+    } else if (m_renderMode == Inline) {
+        // QSGRenderNode-based rendering
         QQuick3DSGRenderNode *n = static_cast<QQuick3DSGRenderNode *>(node);
         if (!n) {
             if (!m_renderNode)
-                m_renderNode = new QQuick3DSGRenderNode();
+                m_renderNode = new QQuick3DSGRenderNode;
             n = m_renderNode;
         }
 
         if (!n->renderer) {
             n->window = window();
             n->renderer = createRenderer();
-            n->renderer->data = n;
-            doImportShaderCache();
         }
 
         const QSize targetSize = window()->effectiveDevicePixelRatio() * QSize(width(), height());
 
         if (isVisible() && isComponentComplete()) {
-            n->renderer->synchronize(this, targetSize, false);
+            n->renderer->synchronize(this, targetSize, window()->effectiveDevicePixelRatio(), false);
             updateDynamicTextures();
             n->markDirty(QSGNode::DirtyMaterial);
         }
 
         return n;
+    } else {
+        qWarning("Invalid renderMode %d", int(m_renderMode));
+        return nullptr;
     }
 }
 
@@ -531,7 +478,8 @@ void QQuick3DViewport::setCamera(QQuick3DCamera *camera)
         return;
 
     m_camera = camera;
-    m_camera->updateGlobalVariables(QRect(0, 0, width(), height()));
+    if (m_camera)
+        m_camera->updateGlobalVariables(QRect(0, 0, width(), height()));
     emit cameraChanged();
     update();
 }
@@ -565,31 +513,32 @@ void QQuick3DViewport::setImportScene(QQuick3DNode *inScene)
             return;
         }
 
-        QQuick3DSceneRootNode *rn = dynamic_cast<QQuick3DSceneRootNode *>(scene);
+        QQuick3DSceneRootNode *rn = qobject_cast<QQuick3DSceneRootNode *>(scene);
         scene = rn ? rn->view3D()->importScene() : nullptr;
     }
 
     m_importScene = inScene;
     if (m_importScene) {
-        // If the referenced scene doesn't have a manager, add one (scenes defined outside of an view3d)
+        // If the referenced scene doesn't have a manager use the one from the sceen root (scenes defined outside of an view3d)
         auto privateObject = QQuick3DObjectPrivate::get(m_importScene);
-        // ### BUG: This will probably leak, need to think harder about this
         if (!privateObject->sceneManager) {
-            QSharedPointer<QQuick3DSceneManager> manager(new QQuick3DSceneManager(m_importScene));
-            manager->setWindow(window());
-            privateObject->refSceneManager(manager);
+            if (QQuick3DSceneManager *manager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager) {
+                manager->setWindow(window());
+                privateObject->refSceneManager(*manager);
+            }
+            Q_ASSERT(privateObject->sceneManager);
         }
 
-        connect(privateObject->sceneManager.data(), &QQuick3DSceneManager::needsUpdate,
+        connect(privateObject->sceneManager, &QQuick3DSceneManager::needsUpdate,
                 this, &QQuickItem::update);
 
         QQuick3DNode *scene = inScene;
         while (scene) {
-            QQuick3DSceneRootNode *rn = dynamic_cast<QQuick3DSceneRootNode *>(scene);
+            QQuick3DSceneRootNode *rn = qobject_cast<QQuick3DSceneRootNode *>(scene);
             scene = rn ? rn->view3D()->importScene() : nullptr;
 
             if (scene) {
-                connect(QQuick3DObjectPrivate::get(scene)->sceneManager.data(),
+                connect(QQuick3DObjectPrivate::get(scene)->sceneManager,
                         &QQuick3DSceneManager::needsUpdate,
                         this, &QQuickItem::update);
             }
@@ -615,12 +564,16 @@ void QQuick3DViewport::setRenderMode(QQuick3DViewport::RenderMode renderMode)
 /*!
     \qmlmethod vector3d View3D::mapFrom3DScene(vector3d scenePos)
 
-    Transforms \a scenePos from scene space (3D) into view space (2D). The
-    returned x- and y-values will be be in view coordinates. The returned z-value
-    will contain the distance from the near side of the frustum (clipNear) to
-    \a scenePos in scene coordinates. If the distance is negative, the point is behind the camera.
-    If \a scenePos cannot be mapped to a position in the scene, a position of [0, 0, 0] is returned.
-    This function requires that a camera is assigned to the view.
+    Transforms \a scenePos from scene space (3D) into view space (2D).
+
+    The returned x- and y-values will be be in view coordinates, with the top-left
+    corner at [0, 0] and the bottom-right corner at [width, height]. The returned
+    z-value contains the distance from the near clip plane of the frustum (clipNear) to
+    \a scenePos in scene coordinates. If the distance is negative, that means the \a scenePos
+    is behind the active camera. If \a scenePos cannot be mapped to a position in the scene,
+    a position of [0, 0, 0] is returned.
+
+    This function requires that \l camera is assigned to the view.
 
     \sa mapTo3DScene(), {Camera::mapToViewport()}{Camera.mapToViewport()}
 */
@@ -643,11 +596,17 @@ QVector3D QQuick3DViewport::mapFrom3DScene(const QVector3D &scenePos) const
 /*!
     \qmlmethod vector3d View3D::mapTo3DScene(vector3d viewPos)
 
-    Transforms \a viewPos from view space (2D) into scene space (3D). The x- and
-    y-values of \a viewPos should be in view coordinates. The z-value should be
-    the distance from the near side of the frustum (clipNear) into the scene in scene
-    coordinates. If \a viewPos cannot be mapped to a position in the scene, a
-    position of [0, 0, 0] is returned.
+    Transforms \a viewPos from view space (2D) into scene space (3D).
+
+    The x- and y-values of \a viewPos should be in view coordinates, with the top-left
+    corner at [0, 0] and the bottom-right corner at [width, height]. The z-value is
+    interpreted as the distance from the near clip plane of the frustum (clipNear) in
+    scene coordinates.
+
+    If \a viewPos cannot successfully be mapped to a position in the scene, a position of
+    [0, 0, 0] is returned.
+
+    This function requires that a \l camera is assigned to the view.
 
     \sa mapFrom3DScene(), {Camera::mapFromViewport}{Camera.mapFromViewport()}
 */
@@ -670,8 +629,13 @@ QVector3D QQuick3DViewport::mapTo3DScene(const QVector3D &viewPos) const
 /*!
     \qmlmethod PickResult View3D::pick(float x, float y)
 
-    Transforms the screen space coordinates \a x and \a y to a ray cast towards that position
-    in scene space. Returns information about the ray hit.
+    This method will "shoot" a ray into the scene from view coordinates \a x and \a y
+    and return information about the nearest intersection with an object in the scene.
+
+    This can, for instance, be called with mouse coordinates to find the object under the mouse cursor.
+
+    \note For an object with a custom \l Geometry, intersections will be approximated based on its bounding
+    box.
 */
 QQuick3DPickResult QQuick3DViewport::pick(float x, float y) const
 {
@@ -728,11 +692,12 @@ QQuick3DSceneRenderer *QQuick3DViewport::getRenderer() const
 void QQuick3DViewport::updateDynamicTextures()
 {
     // Update QSGDynamicTextures that are used for source textures and Quick items
-    // TODO: could be optimized to not update textures that aren't used or are on culled
-    // geometry.
+    // Must be called on the render thread.
+
     const auto &sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
     for (auto *texture : qAsConst(sceneManager->qsgDynamicTextures))
         texture->updateTexture();
+
     QQuick3DNode *scene = m_importScene;
     while (scene) {
         const auto &importSm = QQuick3DObjectPrivate::get(scene)->sceneManager;
@@ -742,7 +707,7 @@ void QQuick3DViewport::updateDynamicTextures()
         }
 
         // if importScene has another import
-        QQuick3DSceneRootNode *rn = dynamic_cast<QQuick3DSceneRootNode *>(scene);
+        QQuick3DSceneRootNode *rn = qobject_cast<QQuick3DSceneRootNode *>(scene);
         scene = rn ? rn->view3D()->importScene() : nullptr;
     }
 }
@@ -754,23 +719,16 @@ void QQuick3DViewport::setupDirectRenderer(RenderMode mode)
     if (!m_directRenderer) {
         m_directRenderer = new QQuick3DSGDirectRenderer(createRenderer(), window(), renderMode);
         connect(window(), &QQuickWindow::sceneGraphInvalidated, this, &QQuick3DViewport::cleanupDirectRenderer, Qt::DirectConnection);
-        doImportShaderCache();
     }
+
     const QSizeF targetSize = window()->effectiveDevicePixelRatio() * QSizeF(width(), height());
     m_directRenderer->setViewport(QRectF(window()->effectiveDevicePixelRatio() * mapToScene(QPointF(0, 0)), targetSize));
     m_directRenderer->setVisibility(isVisible());
     if (isVisible()) {
-        m_directRenderer->renderer()->synchronize(this, targetSize.toSize(), false);
+        m_directRenderer->renderer()->synchronize(this, targetSize.toSize(), window()->effectiveDevicePixelRatio(), false);
         updateDynamicTextures();
         m_directRenderer->requestRender();
     }
-    updateClearBeforeRendering();
-}
-
-void QQuick3DViewport::updateClearBeforeRendering()
-{
-    // Don't clear window when rendering visible underlay
-    window()->setClearBeforeRendering(m_renderMode != Underlay || !isVisible());
 }
 
 // This is used for offscreen mode since we need to check if
